# -*- coding: utf-8 -*-

from twisted.internet.protocol import Protocol, Factory
from netkit.box import Box
import json

from ..utils import safe_call
from ..log import logger
from .. import constants


class AdminConnectionFactory(Factory):

    def __init__(self, proxy):
        self.proxy = proxy

    def buildProtocol(self, addr):
        return AdminConnection(self, addr)


class AdminConnection(Protocol):
    _read_buffer = None

    # 客户端IP的数字
    _client_ip_num = None

    def __init__(self, factory, address):
        self.factory = factory
        self.address = address
        self._read_buffer = ''

    def dataReceived(self, data):
        """
        当数据接受到时
        :param data:
        :return:
        """
        self._read_buffer += data

        while self._read_buffer:
            # 因为box后面还是要用的
            box = Box()
            ret = box.unpack(self._read_buffer)
            if ret == 0:
                # 说明要继续收
                return
            elif ret > 0:
                # 收好了
                self._read_buffer = self._read_buffer[ret:]
                safe_call(self._on_read_complete, box)
                continue
            else:
                # 数据已经混乱了，全部丢弃
                logger.error('buffer invalid. ret: %d, read_buffer: %r', ret, self._read_buffer)
                self._read_buffer = ''
                return

    def _auth_user(self, username, password):
        """
        验证用户
        :param username:
        :param password:
        :return:
        """

        return (self.factory.proxy.app.admin_username or '', self.factory.proxy.app.admin_password or '') == (
            username or '', password or ''
        )

    def _on_read_complete(self, box):
        """
        完整数据接收完成
        :param box: 解析之后的box
        :return:
        """

        # 无论是哪一种请求，都先验证用户
        req_body = json.loads(box.body)

        rsp = None

        if not self._auth_user(req_body['auth']['username'], req_body['auth']['password']):
            rsp = box.map(dict(
                ret=constants.RET_ADIMN_AUTH_FAIL
            ))
        else:
            if box.cmd == constants.CMD_ADMIN_SERVER_STAT:
                idle_workers = dict([(group_id, len(workers)) for group_id, workers in
                                     self.factory.proxy.task_dispatcher.idle_workers_dict.items()])
                busy_workers = dict([(group_id, len(workers)) for group_id, workers in
                                     self.factory.proxy.task_dispatcher.busy_workers_dict.items()])

                # 正在处理的tasks
                pending_tasks = dict([(group_id, queue.qsize()) for group_id, queue in
                                     self.factory.proxy.task_dispatcher.group_queue.queue_dict.items()])

                rsp_body = dict(
                    clients=self.factory.proxy.stat_counter.clients,
                    client_req=self.factory.proxy.stat_counter.client_req,
                    client_rsp=self.factory.proxy.stat_counter.client_rsp,
                    worker_req=self.factory.proxy.stat_counter.worker_req,
                    worker_rsp=self.factory.proxy.stat_counter.worker_rsp,
                    idle_workers=idle_workers,
                    busy_workers=busy_workers,
                    pending_tasks=pending_tasks,
                    tasks_time=dict(self.factory.proxy.stat_counter.tasks_time_counter),
                )

                rsp = box.map(dict(
                    body=json.dumps(rsp_body)
                ))

        if rsp:
            self.transport.write(rsp.pack())




# -*- coding: utf-8 -*-

NAME = 'burst'

# 系统返回码
# 命令字不合法
RET_INVALID_CMD = -10000
# 系统内部异常
RET_INTERNAL = -10001
# admin用户验证失败
RET_ADIMN_AUTH_FAIL = -20000

# 内部使用的命令字
# 分配任务. 如果body里面带数据，说明是要写回；如果没有数据，说明只是要分配task
CMD_WORKER_TASK_ASSIGN = 100
# 任务完成
CMD_WORKER_TASK_DONE = 200
# 管理员命令
# 获取运行状态统计
CMD_ADMIN_SERVER_STAT = 20000


# 子进程城的环境变量
CHILD_ENV_KEY = 'BURST_ENV'

# 作业时间统计标准
TASKS_TIME_BENCHMARK = (10, 50, 100, 500, 1000, 5000)

# proxy<->worker之间通信的address模板
IPC_ADDRESS_TPL = '%s_ipc/' % NAME + '%s.sock'


# 默认backlog
PROXY_BACKLOG = 256


# worker的状态
WORKER_STATUS_IDLE = 1
WORKER_STATUS_BUSY = 2

# worker重连等待时间
WORKER_TRY_CONNECT_INTERVAL = 1

# 网络连接超时(秒)，包括 connect once，read once，write once
WORKER_CONN_TIMEOUT = 3


# 进程类型
PROC_TYPE_MASTER = 'master'
PROC_TYPE_PROXY = 'proxy'
PROC_TYPE_WORKER = 'worker'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import time
import argparse
from collections import OrderedDict
from netkit.box import Box
from netkit.contrib.tcp_client import TcpClient
from burst import constants


class BurstCtl(object):

    args = None

    tcp_client = None

    def __init__(self, args):
        self.args = args

    def make_send_box(self, cmd, username, password):
        return Box(dict(
            cmd=cmd,
            body=json.dumps(
                dict(
                    auth=dict(
                        username=username,
                        password=password,
                    )
                )
            )
        ))

    def output(self, s):
        print '/' + '-' * 80
        print s
        print '-' * 80 + '/'

    def process_stat(self):
        send_box = self.make_send_box(constants.CMD_ADMIN_SERVER_STAT, self.args.username, self.args.password)
        self.tcp_client.write(send_box)

        rsp_box = self.tcp_client.read()

        if not rsp_box:
            self.output('disconnected.')
            return False

        if rsp_box.ret != 0:
            self.output('fail. rsp_box.ret=%s' % rsp_box.ret)
            return False

        body_dict = json.loads(rsp_box.body)

        output_items = []
        for key in ('clients', 'busy_workers', 'idle_workers', 'pending_tasks',
                    'client_req', 'client_rsp', 'worker_req', 'worker_rsp'):

            output_items.append((key, body_dict.get(key)))

        def tasks_time_cmp_func(item1, item2):
            k1 = item1[0]
            k2 = item2[0]
            if k1 == 'more':
                return 1
            if k2 == 'more':
                return -1

            return cmp(int(k1), int(k2))

        tasks_time_items = sorted(body_dict['tasks_time'].items(), cmp=tasks_time_cmp_func)

        output_items.append(('tasks_time', OrderedDict(tasks_time_items)))

        output_dict = OrderedDict(output_items)

        # OrderedDict在通过json打印的时候，会保持原来的顺序
        self.output(json.dumps(output_dict, indent=4))

        return True

    def run(self):
        self.tcp_client = TcpClient(Box, self.args.host, self.args.port, self.args.timeout)

        try:
            self.tcp_client.connect()
        except Exception, e:
            self.output('connect fail: %s' % e)
            return False

        loop_times = 0

        while True:

            result = False
            if self.args.cmd == 'stat':
                result = self.process_stat()

            if not result:
                break

            loop_times += 1
            if loop_times >= self.args.loop > 0:
                break

            try:
                time.sleep(1)
            except KeyboardInterrupt:
                break

        return True


def build_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--host', help='burst admin host', action='store', default='127.0.0.1')
    parser.add_argument('-P', '--port', type=int, help='burst admin port', action='store', required=True)
    parser.add_argument('-u', '--username', help='username', action='store', default=None)
    parser.add_argument('-p', '--password', help='password', action='store', default=None)
    parser.add_argument('-c', '--cmd', help='cmd', action='store', default='stat', choices=('stat',))
    parser.add_argument('-o', '--timeout', type=int, help='connect/send/receive timeout', action='store', default=10)
    parser.add_argument('-l', '--loop', type=int, help='loop times, <=0 means infinite loop', action='store', default=-1)
    return parser


def send_and_recv(tcp_client, box):
    tcp_client.write(box)

    rsp_box = tcp_client.read()

    if not rsp_box:
        print 'disconnected.'
        return False

    if rsp_box.ret != 0:
        print 'fail. rsp_box.ret=%s' % rsp_box.ret
        return False
    else:
        print '/' + '-' * 80
        print json.dumps(json.loads(rsp_box.body), indent=4)
        print '-' * 80 + '/'
        return True


def main():
    args = build_parser().parse_args()
    ctl = BurstCtl(args)
    ctl.run()

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

import signal
import os
# linux 默认就是epoll
from twisted.internet import reactor
import setproctitle

from ..log import logger
from client_connection import ClientConnectionFactory
from worker_connection import WorkerConnectionFactory
from admin_connection import AdminConnectionFactory
from task_dispatcher import TaskDispatcher
from stat_counter import StatCounter
from .. import constants


class Proxy(object):
    """
    proxy相关
    """

    type = constants.PROC_TYPE_PROXY

    client_connection_factory_class = ClientConnectionFactory
    worker_connection_factory_class = WorkerConnectionFactory
    admin_connection_factory_class = AdminConnectionFactory

    app = None

    host = None
    port = None

    # 任务调度器
    task_dispatcher = None
    # 统计
    stat_counter = None

    def __init__(self, app, host, port):
        """
        构造函数
        :return:
        """
        self.app = app
        self.host = host
        self.port = port

        self.task_dispatcher = TaskDispatcher()
        self.stat_counter = StatCounter(self.app.tasks_time_benchmark)

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(self.type))

        # 主进程
        self._handle_proc_signals()

        ipc_directory = os.path.dirname(self.app.ipc_address_tpl)
        if not os.path.exists(ipc_directory):
            os.makedirs(ipc_directory)

        # 启动监听worker
        for group_id in self.app.group_conf:
            ipc_address = self.app.ipc_address_tpl % group_id

            # 防止之前异常导致遗留
            if os.path.exists(ipc_address):
                os.remove(ipc_address)

            # 给内部worker通信用的
            reactor.listenUNIX(ipc_address, self.worker_connection_factory_class(self, group_id))

        # 启动对外监听
        reactor.listenTCP(self.port, self.client_connection_factory_class(self),
                          backlog=self.app.proxy_backlog, interface=self.host)

        # 启动admin服务
        if self.app.admin_address:
            reactor.listenTCP(self.app.admin_address[1], self.admin_connection_factory_class(self),
                              interface=self.app.admin_address[0])

        try:
            reactor.run(installSignalHandlers=False)
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _handle_proc_signals(self):
        def exit_handler(signum, frame):
            """
            在centos6下，callFromThread(stop)无效，因为处理不够及时
            """
            try:
                reactor.stop()
            except:
                pass

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        # 直接停止
        signal.signal(signal.SIGTERM, exit_handler)
        # 忽略，因为这个时候是在重启worker
        signal.signal(signal.SIGHUP, signal.SIG_IGN)



# -*- coding: utf-8 -*-

import socket
from netkit.stream import Stream


class UnixClient(object):
    """
    封装过的tcp client
    """

    box_class = None
    stream_checker = None
    address = None
    timeout = None
    stream = None

    def __init__(self, box_class, address, timeout=None):
        self.box_class = box_class
        self.address = address
        self.timeout = timeout
        self.stream_checker = self.box_class().check

    def connect(self):
        """
        连接服务器，失败会抛出异常
        :return:
        """
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.timeout)
        try:
            sock.connect(self.address)
        except Exception, e:
            sock.close()
            raise e

        self.stream = Stream(sock)

    def read(self):
        """
        如果超时会抛出异常 socket.timeout
        :return:
        """
        if self.closed():
            return None

        data = self.stream.read_with_checker(self.stream_checker)
        if not data:
            return None

        box = self.box_class()
        box.unpack(data)

        return box

    def write(self, data):
        """
        写入
        :param data:
        :return:    True/False
        """
        if self.closed():
            return False

        if isinstance(data, self.box_class):
            data = data.pack()
        elif isinstance(data, dict):
            data = self.box_class(data).pack()

        return self.stream.write(data)

    def close(self):
        if not self.stream:
            return

        return self.stream.close()

    def closed(self):
        if not self.stream:
            return True

        return self.stream.closed()

    def __str__(self):
        return 'box_class: %s, address: %s, timeout: %s' % (self.box_class, self.address, self.timeout)



# -*- coding: utf-8 -*-

import sys
import os
import json
from collections import Counter
from log import logger
from mixins import RoutesMixin, AppEventsMixin
import constants
from proxy import Proxy
from worker import Worker
from master import Master


class Burst(RoutesMixin, AppEventsMixin):

    # 配置都放到 burst 里，而和proxy或者worker直接相关的类，则放到自己的部分
    ############################## configurable begin ##############################

    # 进程名字
    name = constants.NAME

    box_class = None
    group_conf = None
    group_router = None

    # proxy的backlog
    proxy_backlog = constants.PROXY_BACKLOG

    # worker<->proxy网络连接超时(秒)
    worker_conn_timeout = constants.WORKER_CONN_TIMEOUT
    # 处理task超时(秒). 超过后worker会自杀. None 代表永不超时
    work_timeout = None

    # 停止子进程超时(秒). 使用 TERM 进行停止时，如果超时未停止会发送KILL信号
    stop_timeout = None
    # proxy<->worker之间通信的address模板
    ipc_address_tpl = constants.IPC_ADDRESS_TPL

    # 管理员，可以连接proxy获取数据
    # 管理员访问地址如 (127.0.0.1, 22222)
    admin_address = None
    admin_username = None
    admin_password = None

    # 统计相关
    # 作业时间统计标准
    tasks_time_benchmark = constants.TASKS_TIME_BENCHMARK

    ############################## configurable end   ##############################

    blueprints = None

    def __init__(self, box_class, group_conf, group_router):
        """
        构造函数
        :param box_class: box类
        :param group_conf: 进程配置，格式如下:
            {
                $group_id: {
                    count: 10,
                }
            }
        :param group_router: 通过box路由group_id:
            def group_router(box):
                return group_id
        :return:
        """
        RoutesMixin.__init__(self)
        AppEventsMixin.__init__(self)

        self.box_class = box_class
        self.group_conf = group_conf
        self.group_router = group_router

        self.blueprints = list()

    def register_blueprint(self, blueprint):
        blueprint.register_to_app(self)

    def run(self, host, port):
        self._validate_cmds()

        # 只要没有这个环境变量，就是主进程
        str_burst_env = os.getenv(constants.CHILD_ENV_KEY)

        if not str_burst_env:
            # 主进程
            logger.info('Running server on %s:%s', host, port)
            if self.admin_address:
                logger.info('Running admin server on %s:%s', self.admin_address[0], self.admin_address[1])
            Master(self).run()
        else:
            burst_env = json.loads(str_burst_env)
            if burst_env['type'] == constants.PROC_TYPE_PROXY:
                # proxy
                Proxy(self, host, port).run()
            else:
                # worker
                Worker(self, burst_env['group_id']).run()

    def make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s %s:%s] %s' % (
            self.name,
            constants.NAME,
            subtitle,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name

    def _validate_cmds(self):
        """
        确保 cmd 没有重复
        :return:
        """

        cmd_list = list(self.rule_map.keys())

        for bp in self.blueprints:
            cmd_list.extend(bp.rule_map.keys())

        duplicate_cmds = (Counter(cmd_list) - Counter(set(cmd_list))).keys()

        assert not duplicate_cmds, 'duplicate cmds: %s' % duplicate_cmds




# -*- coding: utf-8 -*-

"""
主要为了支持到worker能知道要给返回数据的问题
"""

import weakref
from ..log import logger


class Task(object):

    # 封装好的task_box
    task_box = None

    # 客户端连接的弱引用
    _client_conn_ref = None

    def __init__(self, task_box, client_conn):
        self.task_box = task_box
        self.client_conn = client_conn

    @property
    def client_conn(self):
        # 如果已经释放, ()会返回None
        return self._client_conn_ref()

    @client_conn.setter
    def client_conn(self, value):
        def on_del(reference):
            # 已验证，会删除
            logger.error('reference: %s deleted', reference)

        self._client_conn_ref = weakref.ref(value)



# -*- coding: utf-8 -*-

import os
import copy
import json
import sys
import subprocess
import time
import signal
import setproctitle
from .. import constants


class Master(object):
    """
    master相关
    """

    type = constants.PROC_TYPE_MASTER

    app = None

    enable = True

    # 子进程列表
    processes = None

    def __init__(self, app):
        """
        构造函数
        :return:
        """
        self.app = app
        self.processes = []

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(self.type))

        self._handle_proc_signals()

        self._spawn_workers()

    def _spawn_workers(self):
        def start_child_process(proc_env):
            # 要传入group_id
            worker_env = copy.deepcopy(os.environ)
            worker_env.update({
                constants.CHILD_ENV_KEY: json.dumps(proc_env)
            })

            args = [sys.executable] + sys.argv
            inner_p = subprocess.Popen(args, env=worker_env)
            inner_p.proc_env = proc_env
            return inner_p

        proc_env = dict(
            type=constants.PROC_TYPE_PROXY
        )
        p = start_child_process(proc_env)
        self.processes.append(p)

        for group_id, group_info in self.app.group_conf.items():
            proc_env = dict(
                type=constants.PROC_TYPE_WORKER,
                group_id=group_id,
            )

            # 进程个数
            for it in xrange(0, group_info['count']):
                p = start_child_process(proc_env)
                self.processes.append(p)

        while 1:
            for idx, p in enumerate(self.processes):
                if p and p.poll() is not None:
                    # 说明退出了
                    proc_env = p.proc_env
                    self.processes[idx] = None

                    if self.enable:
                        # 如果还要继续服务
                        p = start_child_process(proc_env)
                        self.processes[idx] = p

            if not filter(lambda x: x, self.processes):
                # 没活着的了
                break

            # 时间短点，退出的快一些
            time.sleep(0.1)

    def _handle_proc_signals(self):
        def exit_handler(signum, frame):
            self.enable = False

            # 如果是终端直接CTRL-C，子进程自然会在父进程之后收到INT信号，不需要再写代码发送
            # 如果直接kill -INT $parent_pid，子进程不会自动收到INT
            # 所以这里可能会导致重复发送的问题，重复发送会导致一些子进程异常，所以在子进程内部有做重复处理判断。
            for p in self.processes:
                if p:
                    p.send_signal(signum)

            # https://docs.python.org/2/library/signal.html#signal.alarm
            if self.app.stop_timeout is not None:
                signal.alarm(self.app.stop_timeout)

        def final_kill_handler(signum, frame):
            if not self.enable:
                # 只有满足了not enable，才发送term命令
                for p in self.processes:
                    if p:
                        p.send_signal(signal.SIGKILL)

        def safe_stop_handler(signum, frame):
            """
            等所有子进程结束，父进程也退出
            """
            self.enable = False

            for p in self.processes:
                if p:
                    p.send_signal(signal.SIGTERM)

            if self.app.stop_timeout is not None:
                signal.alarm(self.app.stop_timeout)

        def safe_reload_handler(signum, frame):
            """
            让所有子进程重新加载
            """
            for p in self.processes:
                if p:
                    p.send_signal(signal.SIGHUP)

        # INT, QUIT为强制结束
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGQUIT, exit_handler)
        # TERM为安全结束
        signal.signal(signal.SIGTERM, safe_stop_handler)
        # HUP为热更新
        signal.signal(signal.SIGHUP, safe_reload_handler)
        # 最终判决，KILL掉子进程
        signal.signal(signal.SIGALRM, final_kill_handler)



# -*- coding: utf-8 -*-

import functools
from .log import logger


def safe_call(func, *args, **kwargs):
    """
    安全调用
    """
    try:
        return func(*args, **kwargs)
    except Exception, e:
        logger.error('exc occur. e: %s, func: %s', e, func, exc_info=True)
        # 调用方可以通过 isinstance(e, BaseException) 来判断是否发生了异常
        return e


def safe_func(func):
    """
    把函数变为安全的
    """
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        return safe_call(func, *args, **kwargs)
    return func_wrapper


def ip_int_to_str(ip_num):
    """
    转换ip格式，从int转为str
    :param ip_num:
    :return:
    """
    import socket
    import struct
    return socket.inet_ntoa(struct.pack("!I", ip_num))


def ip_str_to_int(ip_str):
    """
    转化ip格式，从str转为int
    :param ip_str:
    :return:
    """
    import socket
    import struct

    return struct.unpack("!I", socket.inet_aton(ip_str))[0]



# -*- coding: utf-8 -*-


from ..proxy import TaskBox
from .. import constants
from ..log import logger
from ..utils import ip_int_to_str


class Request(object):
    """
    请求
    """

    # 业务层需要访问的对象，消息内容
    box = None

    # connection
    conn = None
    # 封装的任务box，外面不需要理解
    session = None
    is_valid = False
    blueprint = None
    route_rule = None
    # 是否中断处理，即不调用view_func，主要用在before_request中
    interrupted = False
    # 中断后要写入的数据
    interrupt_data = None

    def __init__(self, conn, task_box):
        self.conn = conn
        self.session = task_box
        # 赋值
        self.is_valid = self._parse_raw_data()

    def _parse_raw_data(self):
        if not self.session.body:
            return True

        try:
            self.box = self.app.box_class()
        except Exception, e:
            logger.error('parse raw_data fail. e: %s, request: %s', e, self)
            return False

        if self.box.unpack(self.session.body) > 0:
            self._parse_route_rule()
            return True
        else:
            logger.error('unpack fail. request: %s', self)
            return False

    def _parse_route_rule(self):
        if self.cmd is None:
            return

        route_rule = self.app.get_route_rule(self.cmd)
        if route_rule:
            # 在app层，直接返回
            self.route_rule = route_rule
            return

        for bp in self.app.blueprints:
            route_rule = bp.get_route_rule(self.cmd)
            if route_rule:
                self.blueprint = bp
                self.route_rule = route_rule
                break

    @property
    def worker(self):
        return self.conn.worker

    @property
    def app(self):
        return self.worker.app

    @property
    def client_ip(self):
        """
        客户端连接IP，外面不需要了解task_box
        :return:
        """
        return ip_int_to_str(self.session.client_ip_num)

    @property
    def cmd(self):
        try:
            return self.box.cmd
        except:
            return None

    @property
    def view_func(self):
        return self.route_rule['view_func'] if self.route_rule else None

    @property
    def endpoint(self):
        if not self.route_rule:
            return None

        bp_endpoint = self.route_rule['endpoint']

        return '.'.join([self.blueprint.name, bp_endpoint] if self.blueprint else [bp_endpoint])

    def write(self, data):
        """
        写回响应，业务代码中请勿调用
        如果处理函数没有return数据的话，data可能为None，此时相当于直接进行ask_for_task
        :param data: 可以是dict也可以是box
        :return:
        """

        if isinstance(data, self.app.box_class):
            data = data.pack()
        elif isinstance(data, dict):
            data = self.box.map(data).pack()

        task_box = TaskBox(dict(
            cmd=constants.CMD_WORKER_TASK_DONE,
            body=data or '',
        ))

        return self.conn.write(task_box.pack())

    def interrupt(self, data=None):
        """
        中断处理
        不能在这里直接write数据，是因为write之后就会告知proxy申请task，而业务层很可能误调用多次
        :param data: 要响应的数据，不传即不响应。多次调用，以最后一次为准。
        :return:
        """
        self.interrupted = True
        self.interrupt_data = data

    def __repr__(self):
        return 'cmd: %r, endpoint: %s, box: %r' % (self.cmd, self.endpoint, self.box)
